from numpy import ndarray
import numpy as np
from typing import List
from rank2plan import Pair


def compute_X_tilde(X: ndarray, pairs: List[Pair]) -> ndarray:
    """Compute the X_tilde matrix.

    Args:
        X (ndarray): The original feature matrix, shape (n_samples, n_features).
        pairs (List[Pair]): The list of pairs, shape (n_pairs,).

    Returns:
        ndarray: The X_tilde matrix, shape (n_pairs, n_features).
    """
    X_tilde = []
    for pair in pairs:
        X_tilde.append(pair.sample_weight * (X[pair.j] - X[pair.i]))
    return np.array(X_tilde)



from abc import ABC, abstractmethod
from typing import List, Optional
from rank2plan.types import Pair
from numpy import ndarray


class Model(ABC):
    @abstractmethod
    def fit(self, X: ndarray, pairs: List[Pair]) -> None:
        """Fit the model to the data.

        Args:
            X (ndarray): The data matrix, shape (n_samples, n_features).

            pairs (List[Pair]): The list of ranking pairs, shape (n_pairs,).
        """
        pass

    @abstractmethod
    def predict(self, X: ndarray) -> ndarray:
        """Pointwise evaluation of the model, where lower values are better.

        Args:
            X (ndarray): The data matrix, shape (n_samples, n_features).

        Returns:
            npt.ArrayLike: The pointwise evaluation of the model, shape
            (n_samples,).
        """
        pass

    @abstractmethod
    def weights(self) -> Optional[ndarray]:
        """Get the weights of the model.

        Returns:
            ndarray: The weights of the model, shape (n_features,). None if the
            model has not been fitted.
        """
        pass



from typing import List, Optional
from rank2plan.model import Model
from rank2plan.types import Pair
import numpy as np
from numpy import ndarray
from pulp import (
    LpProblem,
    LpMinimize,
    LpVariable,
    lpDot,
    lpSum,
    LpAffineExpression,
    LpSolver,
)
from dataclasses import dataclass
import logging

LOGGER = logging.getLogger(__name__)


@dataclass
class FitState:
    """A very primitive state, instead of saving the built problem, we just
    rebuild each time, since PrimalLpModel is mainly for debugging and testing
    anyway.
    """

    X: ndarray
    pairs: List[Pair]


class PrimalLpModel(Model):
    """The primal LP model for RankSVM. Based on equation (5) of (Dedieu et al,
    2022). This should mainly be used for debugging and testing purposes.
    """

    def __init__(
        self,
        solver: LpSolver,
        C=1.0,
    ) -> None:
        """Initialise the primal LP model.

        Args:
            solver (LpSolver): The solver to use.
            C (float, optional): Regularisation parameter. Defaults to 1.0.
        """
        self.solver = solver
        self.C = C
        self.state: Optional[FitState] = None
        self._weights = None

    def fit(self, X: ndarray, pairs: List[Pair], save_state=False) -> None:
        N = X.shape[0]
        P = X.shape[1]

        prob = LpProblem("PrimalLp", LpMinimize)
        beta_plus: List[LpVariable] = []
        beta_minus: List[LpVariable] = []

        for p in range(P):
            beta_plus.append(LpVariable(f"beta_plus_{p}", lowBound=0))
            beta_minus.append(LpVariable(f"beta_minus_{p}", lowBound=0))

        LOGGER.info("Created variables, constructing h values")
        LOGGER.info(f"Constructing h values, building problem")

        xi: List[LpVariable] = []
        sample_weights: List[float] = []
        for pair_id, pair in enumerate(pairs):
            h_value_i = lpDot(beta_plus, X[pair.i]) - lpDot(beta_minus, X[pair.i])
            h_value_j = lpDot(beta_plus, X[pair.j]) - lpDot(beta_minus, X[pair.j])
            xi.append(LpVariable(f"xi_{pair_id}_{pair.i}_{pair.j}", lowBound=0))
            # this order of j and i makes lower scores better
            prob += h_value_j - h_value_i >= pair.gap - xi[-1]
            sample_weights.append(pair.sample_weight)

        main_objective: LpAffineExpression = lpDot(xi, sample_weights)  # type: ignore
        regularisation_objective: LpAffineExpression = lpSum(beta_plus) + lpSum(
            beta_minus
        )
        prob += main_objective * self.C + regularisation_objective
        LOGGER.info("Problem build, solving")

        prob.solve(self.solver)

        LOGGER.info("Finished solving")
        LOGGER.info(f"Overall objective: {prob.objective.value()}")  # type: ignore
        LOGGER.info(f"Main objective: {main_objective.value()}")
        LOGGER.info(f"Regularisation objective: {regularisation_objective.value()}")

        self._weights = np.array(
            [beta_plus[i].varValue - beta_minus[i].varValue for i in range(P)]  # type: ignore
        )
        if save_state:
            self.state = FitState(X, pairs)

    def refit_with_C_value(self, C: float, save_state=False) -> None:
        assert self.state is not None
        LOGGER.info(f"Refitting with C value changed from {self.C} to {C}")
        self.C = C
        self.fit(self.state.X, self.state.pairs, save_state=save_state)

    def predict(self, X: ndarray) -> ndarray:
        return X @ self._weights

    def weights(self) -> Optional[ndarray]:
        return self._weights



from enum import Enum, auto


class Pair:
    def __init__(self, i: int, j: int, gap: float = 1.0, sample_weight: float = 1.0):
        self.i = i
        self.j = j
        self.gap = gap
        self.sample_weight = sample_weight


class LossType(Enum):
    Hinge = auto()
    SquaredHinge = auto()


class PenalisationType(Enum):
    L1 = auto()
    L2 = auto()


class TuningMetric(Enum):
    KendallTau = auto()
    # See documents/theory.pdf for what the LpMainObjective is
    LpMainObjective = auto()



from rank2plan import Model, Pair, LossType, PenalisationType
from rank2plan.lp_models.constraint_column_generation.smoothing_hinge_loss import (
    loop_smoothing_hinge_loss_columns_samples_restricted,
    loop_smoothing_hinge_loss_samples_restricted,
)
from rank2plan.lp_models.objective_values import (
    compute_main_objective,
    compute_regularisation_objective,
    compute_overall_objective,
)
from rank2plan.lp_models.constraint_column_generation.utils import compute_X_tilde
from pulp import LpSolver, LpProblem, LpMinimize, LpVariable, lpSum, lpDot, LpConstraint
from typing import List, Optional, Tuple
from numpy import ndarray
import numpy as np
from time import time
import logging
import random
import math
from dataclasses import dataclass

LOGGER = logging.getLogger(__name__)


@dataclass
class FitState:
    N: int
    P: int
    X_tilde: ndarray
    pairs: List[Pair]
    problem: LpProblem
    constraint_indices: List[int]
    feature_indices: List[int]


class ConstraintColumnModel(Model):
    def __init__(
        self, solver: LpSolver, C: float, tol: float, no_feature_sampling=False
    ) -> None:
        """Don't use this directly, use LpModel instead."""
        if solver.mip == True:
            raise ValueError("solver must be configured for LP, use mip=False")
        self.solver = solver
        self.C = C
        self.tol = tol
        self.no_feature_sampling = no_feature_sampling
        self.state: Optional[FitState] = None
        self._weights: Optional[ndarray] = None

    def fit(self, X: ndarray, pairs: List[Pair], save_state=False) -> None:
        start = time()
        X_tilde = compute_X_tilde(X, pairs)
        LOGGER.info(f"Computed X_tilde in {time() - start:.2f}s")

        start = time()
        if self.no_feature_sampling:
            constraint_indices = _init_constraint_sampling_smoothing(
                X_tilde, pairs, self.C
            )
            feature_indices = list(range(X_tilde.shape[1]))
        else:
            constraint_indices, feature_indices = (
                _init_constraint_column_sampling_smoothing(X_tilde, pairs, self.C)
            )
        LOGGER.info(
            f"Initial constraint and column sampling done in {time() - start:.3f} seconds"
        )

        problem = self._build_subproblem(
            X_tilde, pairs, constraint_indices, feature_indices, None
        )

        N, P = X_tilde.shape
        self.state = FitState(
            N, P, X_tilde, pairs, problem, constraint_indices, feature_indices
        )
        self._fit()
        if not save_state:
            self.state = None

    def refit_with_C_value(self, C: float, save_state=False) -> None:
        assert self.state is not None
        LOGGER.info(f"Refitting with C value changed from {self.C} to {C}")
        self.C = C
        self.state.problem = self._rebuild_subproblem_objective(self.state.problem)
        self._fit()
        if not save_state:
            self.state = None

    def predict(self, X: ndarray) -> ndarray:
        return X @ self._weights

    def weights(self) -> Optional[ndarray]:
        return self._weights

    def _fit(self) -> None:
        assert self.state is not None
        N, P = self.state.N, self.state.P
        features_to_check = list(set(range(P)) - set(self.state.feature_indices))
        constraint_to_check = list(set(range(N)) - set(self.state.constraint_indices))
        gs = np.array([pair.gap * pair.sample_weight for pair in self.state.pairs])

        cur_iter = 0
        while True:
            cur_iter += 1

            start = time()
            self.state.problem.solve(self.solver)
            LOGGER.info(f"Subproblem solved in {time() - start:.3f} seconds")
            LOGGER.info(f"Current objective: {self.state.problem.objective.value()}")  # type: ignore

            if (
                len(self.state.feature_indices) == P
                and len(self.state.constraint_indices) == N
            ):
                LOGGER.info("Finishing as all features and constraints are used")
                break

            constraints: List[LpConstraint] = [
                self.state.problem.constraints[_constraint_name(i)]
                for i in self.state.constraint_indices
            ]
            dual_values = np.array([constraint.pi for constraint in constraints])

            variable_dict = self.state.problem.variablesDict()
            beta_plus = [
                variable_dict[_beta_plus_name(feature_id)]
                for feature_id in self.state.feature_indices
            ]
            beta_minus = [
                variable_dict[_beta_minus_name(feature_id)]
                for feature_id in self.state.feature_indices
            ]
            beta = np.array(
                [
                    beta_plus[i].varValue - beta_minus[i].varValue
                    for i in range(len(self.state.feature_indices))
                ]
            )

            # Reduced costs for features
            X_tilde_reduced = self.state.X_tilde[self.state.constraint_indices, :][
                :, features_to_check
            ]
            reduced_costs_features = (1 / self.C) * np.ones(
                len(features_to_check)
            ) - np.abs(np.dot(X_tilde_reduced.T, dual_values))
            violated_features = np.array(features_to_check)[
                reduced_costs_features < -self.tol
            ]

            if violated_features.shape[0] > 400:
                indices = np.argsort(reduced_costs_features)[:400]
                violated_features = np.array(features_to_check)[indices]

            X_tilde_reduced = self.state.X_tilde[:, self.state.feature_indices][
                constraint_to_check, :
            ]
            reduced_costs_constraints = gs[constraint_to_check] - np.dot(
                X_tilde_reduced, beta
            )
            violated_constraints = np.array(constraint_to_check)[
                reduced_costs_constraints > self.tol
            ]

            if len(violated_constraints) == 0 and len(violated_features) == 0:
                LOGGER.info("Finishing as no violated constraints or features")
                break

            if len(violated_features) > 0:
                LOGGER.info(f"Adding {len(violated_features)} features")
                LOGGER.info(
                    f"Most negative feature reduced cost: {np.min(reduced_costs_features)}"
                )
                self.state.problem = self._add_features_to_subproblem(
                    self.state.X_tilde,
                    self.state.pairs,
                    self.state.problem,
                    violated_features,
                    self.state.constraint_indices,
                )
                self.state.feature_indices += list(violated_features)
                features_to_check = list(
                    set(range(P)) - set(self.state.feature_indices)
                )

                variable_dict = self.state.problem.variablesDict()
                beta_plus += [
                    variable_dict[_beta_plus_name(feature_id)]
                    for feature_id in violated_features
                ]
                beta_minus += [
                    variable_dict[_beta_minus_name(feature_id)]
                    for feature_id in violated_features
                ]

            if len(violated_constraints) > 0:
                LOGGER.info(f"Adding {len(violated_constraints)} constraints")
                LOGGER.info(
                    f"Most violated constraint: {np.max(reduced_costs_constraints)}"
                )
                self.state.problem = self._add_constraints_to_subproblem(
                    self.state.X_tilde,
                    self.state.pairs,
                    self.state.problem,
                    violated_constraints,
                    self.state.feature_indices,
                    beta_plus,
                    beta_minus,
                )
                self.state.constraint_indices.extend(violated_constraints)
                constraint_to_check = list(
                    set(constraint_to_check) - set(violated_constraints)
                )

        variable_dict = self.state.problem.variablesDict()
        beta_plus = np.array(
            [
                (
                    variable_dict[_beta_plus_name(feature_id)].varValue
                    if feature_id in self.state.feature_indices
                    else 0
                )
                for feature_id in range(P)
            ]
        )
        beta_minus = np.array(
            [
                (
                    variable_dict[_beta_minus_name(feature_id)].varValue
                    if feature_id in self.state.feature_indices
                    else 0
                )
                for feature_id in range(P)
            ]
        )
        self._weights = beta_plus - beta_minus
        LOGGER.info(
            f"Finished constraint and column generation in {cur_iter} iterations, using {time() - start:.3f} seconds"
        )
        LOGGER.info(
            f"Final support size {len(self.state.feature_indices)}, of which {len(np.where(self._weights != 0)[0])} are non-zero"
        )
        LOGGER.info(f"Pulp objective: {self.state.problem.objective.value()}")  # type: ignore
        LOGGER.info(
            f"Overall objective: {compute_overall_objective(self.state.X_tilde, self.state.pairs, self._weights, self.C)}"
        )
        LOGGER.info(
            f"Main objective: {compute_main_objective(self.state.X_tilde, self.state.pairs, self._weights)}"
        )
        LOGGER.info(
            f"Regularisation objective: {compute_regularisation_objective(self._weights)}"
        )

    def _build_subproblem(
        self,
        X_tilde: ndarray,
        pairs: List[Pair],
        constraint_indices: List[int],
        feature_indices: List[int],
        warm_start: Optional[ndarray],
    ) -> LpProblem:
        start = time()
        N, P = X_tilde.shape
        N_constraints = len(constraint_indices)
        P_features = len(feature_indices)
        sample_weights = [
            pairs[constraint_id].sample_weight for constraint_id in constraint_indices
        ]
        LOGGER.info(
            f"Subproblem with {N_constraints} constraints and {P_features} features"
        )

        problem = LpProblem("ConstraintColumnGenerationSubproblem", LpMinimize)

        # Hinge loss
        xi = [
            LpVariable(
                _xi_name(constraint_id, pairs[constraint_id].i, pairs[constraint_id].j),
                lowBound=0,
            )
            for constraint_id in constraint_indices
        ]

        # Beta
        beta_plus = [
            LpVariable(_beta_plus_name(feature_id), lowBound=0)
            for feature_id in feature_indices
        ]
        beta_minus = [
            LpVariable(_beta_minus_name(feature_id), lowBound=0)
            for feature_id in feature_indices
        ]

        problem += (
            self.C * lpDot(sample_weights, xi) + lpSum(beta_plus) + lpSum(beta_minus)
        )

        # The selected constraints using the selected features
        X_tilde_reduced = X_tilde[:, feature_indices]
        for i, constraint_id in enumerate(constraint_indices):
            pair = pairs[constraint_id]
            problem.addConstraint(
                (
                    lpDot(X_tilde_reduced[constraint_id], beta_plus)
                    - lpDot(X_tilde_reduced[constraint_id], beta_minus)
                    >= pair.gap * pair.sample_weight - pair.sample_weight * xi[i]
                ),
                name=f"constraint_{constraint_id}",
            )

        if warm_start is not None:
            for i, feature_id in enumerate(feature_indices):
                beta_plus[i].setInitialValue(max(0, warm_start[feature_id]))
                beta_minus[i].setInitialValue(max(0, -warm_start[feature_id]))

        LOGGER.info(f"Subproblem built in {time() - start:.3f} seconds")
        return problem

    def _add_features_to_subproblem(
        self,
        X_tilde: ndarray,
        pairs: List[Pair],
        problem: LpProblem,
        violated_features: ndarray,
        current_constraints: List[int],
    ) -> LpProblem:
        # add new features
        added_beta_plus = []
        added_beta_minus = []

        for feature_id in violated_features:
            added_beta_plus.append(LpVariable(_beta_plus_name(feature_id), lowBound=0))
            added_beta_minus.append(
                LpVariable(_beta_minus_name(feature_id), lowBound=0)
            )

        X_tilde_reduced = X_tilde[:, violated_features]
        problem.addVariables(added_beta_plus)
        problem.addVariables(added_beta_minus)
        for constraint_id in current_constraints:
            old_constraint: LpConstraint = problem.constraints[
                _constraint_name(constraint_id)
            ]
            updated_constraint = (
                old_constraint
                + lpDot(X_tilde_reduced[constraint_id], added_beta_plus)
                - lpDot(X_tilde_reduced[constraint_id], added_beta_minus)
            )
            problem.constraints[_constraint_name(constraint_id)] = updated_constraint
            problem.modifiedConstraints.append(updated_constraint)

        problem.setObjective(
            problem.objective + lpSum(added_beta_plus) + lpSum(added_beta_minus)
        )
        LOGGER.info("Features added and objective updated")

        return problem

    def _add_constraints_to_subproblem(
        self,
        X_tilde: ndarray,
        pairs: List[Pair],
        problem: LpProblem,
        violated_constraints: ndarray,
        current_features: List[int],
        beta_plus: List[LpVariable],
        beta_minus: List[LpVariable],
    ):
        xis_to_add = []
        sample_weights_to_add = []
        X_reduced = X_tilde[:, current_features]

        for violated_constraint in violated_constraints:
            pair = pairs[violated_constraint]
            xi_violated = LpVariable(
                f"xi_{violated_constraint}_{pair.i}_{pair.j}", lowBound=0
            )
            problem.addConstraint(
                (
                    lpDot(X_reduced[violated_constraint], beta_plus)
                    - lpDot(X_reduced[violated_constraint], beta_minus)
                    >= pair.gap * pair.sample_weight - pair.sample_weight * xi_violated
                ),
                name=f"constraint_{violated_constraint}",
            )
            xis_to_add.append(xi_violated)
            sample_weights_to_add.append(pair.sample_weight)

        problem.setObjective(
            problem.objective + lpDot(xis_to_add, sample_weights_to_add) * self.C
        )
        LOGGER.info("Constraints added and objective updated")

        return problem

    def _rebuild_subproblem_objective(self, problem: LpProblem) -> LpProblem:
        assert self.state is not None
        variables_dict = problem.variablesDict()
        beta_plus = [
            variables_dict[_beta_plus_name(feature_id)]
            for feature_id in self.state.feature_indices
        ]
        beta_minus = [
            variables_dict[_beta_minus_name(feature_id)]
            for feature_id in self.state.feature_indices
        ]
        xis = [
            variables_dict[
                _xi_name(
                    constraint_id,
                    self.state.pairs[constraint_id].i,
                    self.state.pairs[constraint_id].j,
                )
            ]
            for constraint_id in self.state.constraint_indices
        ]
        sample_weights = [
            self.state.pairs[constraint_id].sample_weight
            for constraint_id in self.state.constraint_indices
        ]
        problem.setObjective(
            self.C * lpDot(xis, sample_weights) + lpSum(beta_plus) + lpSum(beta_minus)
        )
        LOGGER.info("Rebuilt subproblem objective")
        return problem


def _beta_plus_name(feature_id: int) -> str:
    return f"beta_plus_{feature_id}"


def _beta_minus_name(feature_id: int) -> str:
    return f"beta_minus_{feature_id}"


def _xi_name(constraint_id: int, i: int, j: int) -> str:
    return f"xi_{constraint_id}_{i}_{j}"


def _constraint_name(constraint_id: int) -> str:
    return f"constraint_{constraint_id}"


def _init_constraint_column_sampling_smoothing(
    X_tilde: ndarray, pairs: List[Pair], C: float
) -> Tuple[List[int], List[int]]:
    pairs_np = np.array(pairs)
    N, P = X_tilde.shape
    rho = (1 / C) / np.max(np.sum(np.abs(X_tilde), axis=0))

    N0 = 5 * int(math.sqrt(N))
    P0 = 5 * int(math.sqrt(P))

    if N0 >= N or P0 >= P:
        LOGGER.warning(
            "Using no samples and features for initial sampling, as N or P is too small"
        )
        return [], []

    old_beta_averaged = np.ones(P)
    beta_averaged = np.zeros(P)
    delta_l2_diff_mean = 1e8

    k = 0
    while delta_l2_diff_mean > 5e-1 and k < int(N / N0):
        k += 1
        LOGGER.info(f"Sampling iteration {k}")
        LOGGER.info(f"Difference in l2 norm: {delta_l2_diff_mean}")

        subset = np.sort(random.sample(range(N), N0))
        X_tilde_reduced = X_tilde[subset]
        pairs_reduced = pairs_np[subset]

        # Correlation screening
        argsort_columns = np.argsort(np.abs(np.sum(X_tilde_reduced, axis=0)))
        index_columns = argsort_columns[::-1][:P0]

        X_tilde_reduced = X_tilde_reduced[:, index_columns]

        alpha_sample = rho * np.max(np.sum(np.abs(X_tilde_reduced), axis=0))

        tau_max = 0.2
        n_loop = 20
        n_iter = 100
        beta_smoothing_reduced = loop_smoothing_hinge_loss_columns_samples_restricted(
            LossType.Hinge,
            PenalisationType.L1,
            X_tilde_reduced,
            pairs_reduced,
            alpha_sample,
            tau_max,
            n_loop,
            n_iter,
        )

        beta_sample = np.zeros(P)
        for i, index in enumerate(index_columns):
            beta_sample[index] = beta_smoothing_reduced[i]

        old_beta_averaged = np.copy(beta_averaged)
        beta_averaged += np.array(beta_sample)
        delta_l2_diff_mean = np.linalg.norm(
            1.0 / max(1, k) * beta_averaged - 1.0 / max(1, k - 1) * old_beta_averaged
        )

    beta_averaged *= 1.0 / k

    max_n_cols = 200
    index_columns_smoothing = np.where(beta_averaged != 0)[0]
    LOGGER.info(f"Len support primal: {len(index_columns_smoothing)}")

    if len(index_columns_smoothing) > max_n_cols:
        argsort_columns = np.argsort(np.abs(beta_averaged))
        index_columns_smoothing = argsort_columns[::-1][:max_n_cols]
        LOGGER.info(f"Reduced primal support size to {max_n_cols}")

    gs = np.array([pair.gap * pair.sample_weight for pair in pairs])
    constraints = 1 * gs - np.dot(X_tilde, beta_averaged)
    index_samples_smoothing = np.arange(N)[constraints >= 0]
    LOGGER.info(f"Len dual smoothing: {len(index_samples_smoothing)}")

    return list(index_samples_smoothing), list(index_columns_smoothing)


def _init_constraint_sampling_smoothing(
    X_tilde: ndarray, pairs: List[Pair], C: float, is_restricted=True
) -> List[int]:
    """Initial sampling for constraint generation using smoothing.

    Args:
        X_tilde (ndarray): The X_tilde matrix, shape (n_pairs, n_features).
        pairs (List[Pair]): The list of pairs, shape (n_pairs,).
        C (float): The regularisation parameter.
        is_restricted (bool, optional): Not sure yet :(. Defaults to True.

    Returns:
        List[int]: the initial set of constraints
    """

    pairs_np = np.array(pairs)
    N, P = X_tilde.shape
    rho = (1 / C) / np.max(
        np.sum(np.abs(X_tilde), axis=0)
    )  # lambda (1/C) relative to norms

    N0 = int(min(10 * P, 5 * math.sqrt(N)))
    if N0 < 3:
        LOGGER.warning("Not enough samples to perform initial sampling, returning none")
        return []

    LOGGER.info("Finding initial solution using first-order method")

    tau_max = 0.1
    n_loop = 20
    n_iter = 20

    # result
    old_beta_averaged = -np.ones(P)
    beta_averaged = np.zeros(P)
    delta_variance = 1e6

    k = 0
    while delta_variance > 5e-2 and k < int(N / N0):
        k += 1
        LOGGER.info(f"Sample number: {k}")
        LOGGER.info(f"Difference variance: {delta_variance}")

        subset = np.sort(random.sample(range(N), N0))
        X_tilde_reduced = X_tilde[subset]
        pairs_reduced = pairs_np[subset]

        alpha_sample = rho * np.max(np.sum(np.abs(X_tilde_reduced), axis=0))

        if is_restricted:
            n_loop = 10
            _, beta_sample = loop_smoothing_hinge_loss_samples_restricted(
                LossType.Hinge,
                PenalisationType.L1,
                pairs_reduced,
                X_tilde_reduced,
                alpha_sample,
                tau_max,
                n_loop,
                n_iter,
            )
        else:
            raise NotImplementedError("Unrestricted sampling not implemented yet.")
            # _, _, _, beta_sample, beta0_sample = loop_smoothing_hinge_loss(
            #     LossType.Hinge,
            #     PenalisationType.L1,
            #     X,
            #     pairs_reduced,
            #     alpha_sample,
            #     tau_max,
            #     n_loop,
            #     n_iter,
            # )
            # beta_sample = np.concatenate([beta_sample, np.array([beta0_sample])])

        old_beta_averaged = np.copy(beta_averaged)
        beta_averaged += np.array(beta_sample)
        delta_variance = np.linalg.norm(
            1.0 / max(1, k) * beta_averaged - 1.0 / max(1, k - 1) * old_beta_averaged
        )

    # ---Determine set of constraints
    beta_averaged *= N0 / float(N)

    gs = np.array([pair.gap * pair.sample_weight for pair in pairs])

    constraints = 1 * gs - (np.dot(X_tilde, beta_averaged))
    idx_samples_smoothing = np.arange(N)[constraints >= 0]

    LOGGER.info("Finished initial sampling")
    LOGGER.info(f"Len dual smoothing: {len(idx_samples_smoothing)}")

    return list(idx_samples_smoothing)



import numpy as np
from numpy import ndarray
import time
import math
from rank2plan import Pair, LossType, PenalisationType
import logging

LOGGER = logging.getLogger(__name__)


def smoothing_hinge_loss(
    loss_type: LossType,
    penalisation_type: PenalisationType,
    X_tilde: ndarray,
    pairs: ndarray,
    alpha: float,
    beta_start: ndarray,
    highest_eig: float,
    tau: float,
    n_iter: int,
    is_sparse=False,
):
    # TYPE_PENALIZATION = 1 : L1 -> soft thresholding
    # TYPE_PENALIZATION = 2 : L2

    # ---Initialization
    start_time = time.time()
    N, P = X_tilde.shape

    old_beta = np.ones(P)
    beta_m = beta_start

    # ---MAIN LOOP
    test = 0
    t_AGD_old = 1
    t_AGD = 1
    eta_m_old = beta_start
    g = np.array([pair.gap for pair in pairs])
    s = np.array([pair.sample_weight for pair in pairs])
    gs = np.array([g[i] * s[i] for i in range(N)])

    if loss_type == LossType.Hinge:
        Lipchtiz_coeff = highest_eig / (4 * tau)
    elif loss_type == LossType.SquaredHinge:
        raise ValueError("Squared hinge not implemented")
        # the following line MAY be correct
        # Lipchtiz_coeff = 2 * highest_eig

    while np.linalg.norm(beta_m - old_beta) > 1e-3 and test < n_iter:
        test += 1
        aux = gs - (np.dot(X_tilde, beta_m) if not is_sparse else X_tilde.dot(beta_m))

        # ---Hinge loss
        if loss_type == LossType.Hinge:
            w_tau = [
                min(1, abs(aux[i]) / (2 * tau)) * np.sign(aux[i]) for i in range(N)
            ]
            # gradient_loss   = -0.5*np.sum([y[i]*(1+w_tau[i])*X_add[i,:] for i in range(N)], axis=0)
            gradient_aux = np.array([1 + w_tau[i] for i in range(N)])

            gradient_loss = (
                -0.5 * np.dot(X_tilde.T, gradient_aux)
                if not is_sparse
                else -0.5 * X_tilde.T.dot(gradient_aux)
            )

        # ---Gradient descent
        old_beta = beta_m
        grad = beta_m - 1 / float(Lipchtiz_coeff) * gradient_loss

        # ---Thresholding of top 100 guys !
        dict_thresholding = {
            PenalisationType.L1: soft_thresholding_l1,
            PenalisationType.L2: soft_thresholding_l2,
        }
        eta_m = np.array(
            [
                dict_thresholding[penalisation_type](grad[i], alpha / Lipchtiz_coeff)
                for i in range(P)
            ]
        )

        # ---AGD
        t_AGD = (1 + math.sqrt(1 + 4 * t_AGD_old**2)) / 2.0
        aux_t_AGD = (t_AGD_old - 1) / t_AGD

        beta_m = eta_m + aux_t_AGD * (eta_m - eta_m_old)

        t_AGD_old = t_AGD
        eta_m_old = eta_m

    LOGGER.info(f"Number of iterations: {test}")
    LOGGER.info(f"X_tilde shape: {X_tilde.shape}")

    # ---Support
    idx_columns_smoothing = np.where(beta_m[:P] != 0)[0]
    LOGGER.info(f"Len support smoothing: {idx_columns_smoothing.shape[0]}")

    # ---Constraints
    ##### USE B0 !!!!!!!
    if not is_sparse:
        constraints = 1.05 * gs - (
            np.dot(X_tilde[:, idx_columns_smoothing], beta_m[idx_columns_smoothing])
        )
    else:
        constraints = 1.05 * gs - (
            X_tilde[:, idx_columns_smoothing].dot(beta_m[idx_columns_smoothing])
        )

    idx_samples_smoothing = np.arange(N)[constraints >= 0]

    LOGGER.info(f"Number violated constraints: {idx_samples_smoothing.shape[0]}")
    LOGGER.info(f"Convergence rate: {np.linalg.norm(beta_m - old_beta):.3f}")

    time_smoothing = time.time() - start_time
    LOGGER.info(f"Smoothing time: {time_smoothing:.3f}")

    return (
        idx_samples_smoothing.tolist(),
        idx_columns_smoothing.tolist(),
        time_smoothing,
        beta_m,
    )


# # def loop_smoothing_hinge_loss(
# #     type_loss,
# #     type_penalization,
# #     X,
# #     y,
# #     alpha,
# #     tau_max,
# #     n_loop,
# #     n_iter,
# #     f,
# #     is_sparse=False,
# # ):

# #     # n_loop: how many times should we run the loop ?
# #     # Apply the smoothing technique from the best subset selection

# #     start_time = time.time()
# #     N, P = X.shape
# #     old_beta = -np.ones(P + 1)

# #     # ---New matrix and SVD
# #     if not is_sparse:
# #         X_add = 1 / math.sqrt(N) * np.ones((N, P + 1))
# #         X_add[:, :P] = X
# #         highest_eig = power_method(X_add)
# #     else:
# #         X_add = csr_matrix(hstack([X, coo_matrix(1 / math.sqrt(N) * np.ones((N, 1)))]))
# #         highest_eig = power_method(X_add, is_sparse=True)

# #     beta_smoothing = np.zeros(P + 1)
# #     time_smoothing_sum = 0

# #     tau = tau_max

# #     test = 0
# #     while np.linalg.norm(beta_smoothing - old_beta) > 1e-3 and test < n_loop:
# #         print("TEST CV BEFORE TAU: " + str(np.linalg.norm(beta_smoothing - old_beta)))

# #         test += 1
# #         old_beta = beta_smoothing

# #         idx_samples, idx_columns, time_smoothing, beta_smoothing = smoothing_hinge_loss(
# #             type_loss,
# #             type_penalization,
# #             X,
# #             y,
# #             alpha,
# #             beta_smoothing,
# #             X_add,
# #             highest_eig,
# #             tau,
# #             n_iter,
# #             f,
# #             is_sparse,
# #         )

# #         # ---Update parameters
# #         time_smoothing_sum += time_smoothing
# #         tau = 0.7 * tau

# #     # print beta_smoothing[idx_columns]

# #     time_smoothing_tot = time.time() - start_time
# #     write_and_print("\nNumber of iterations              : " + str(test), f)
# #     write_and_print(
# #         "Total time smoothing for "
# #         + str(type_loss)
# #         + ": "
# #         + str(round(time_smoothing_tot, 3)),
# #         f,
# #     )

# #     return (
# #         idx_samples,
# #         idx_columns,
# #         time_smoothing_sum,
# #         beta_smoothing[:-1],
# #         beta_smoothing[-1],
# #     )


def loop_smoothing_hinge_loss_samples_restricted(
    loss_type: LossType,
    penalisation_type: PenalisationType,
    pairs: ndarray,
    X_tilde: ndarray,
    alpha: float,
    tau_max: float,
    n_loop: int,
    n_iter: int,
):
    # n_loop: how many times should we run the loop ?
    # Apply the smoothing technique from the best subset selection

    start_time = time.time()

    N, P = X_tilde.shape
    old_beta = -np.ones(P)

    # ---New matrix and SVD
    highest_eig = power_method(X_tilde)

    # ---Results
    beta_smoothing = np.zeros(P)
    tau = tau_max

    # ---Prepare for restrcition
    idx_samples = np.arange(N)
    X_tilde_reduced = X_tilde
    pairs_reduced = pairs

    test = -1
    while np.linalg.norm(beta_smoothing - old_beta) > 1e-4 and test < n_loop:
        LOGGER.info(
            f"Test CV before tau: {np.linalg.norm(beta_smoothing - old_beta):.4f}"
        )

        test += 1
        old_beta = beta_smoothing

        idx_samples_restricted, _, _, beta_smoothing = smoothing_hinge_loss(
            loss_type,
            penalisation_type,
            X_tilde_reduced,
            pairs_reduced,
            alpha,
            beta_smoothing,
            highest_eig,
            tau,
            n_iter,
        )

        if test == 0:
            # ---Restrict to samples
            X_tilde_reduced = X_tilde[idx_samples_restricted, :]
            pairs_reduced = pairs[idx_samples_restricted]

            highest_eig = power_method(X_tilde_reduced)
            idx_samples = idx_samples[idx_samples_restricted]

        # ---Update parameters
        tau = 0.7 * tau

    time_smoothing_tot = time.time() - start_time
    LOGGER.info(f"Num iters: {test}")
    LOGGER.info(f"Total smoothing time: {time_smoothing_tot:.3f}")

    return idx_samples.tolist(), beta_smoothing


def loop_smoothing_hinge_loss_columns_samples_restricted(
    loss_type: LossType,
    penalisation_type: PenalisationType,
    X_tilde: ndarray,
    pairs: ndarray,
    alpha,
    tau_max,
    n_loop,
    n_iter,
):

    start_time = time.time()
    N, P = X_tilde.shape
    old_beta = -np.ones(P)

    highest_eig = power_method(X_tilde)

    # ---Results
    beta_smoothing = np.zeros(P)

    tau = tau_max

    # ---Prepare for restriction
    X_tilde_reduced = X_tilde
    pairs_reduced = pairs
    idx_columns_restricted = np.arange(P)

    test = -1
    while np.linalg.norm(beta_smoothing - old_beta) > 1e-2 and test < n_loop:
        LOGGER.info(
            f"Test CV between 2 values of tau: {np.linalg.norm(beta_smoothing - old_beta):.4f}"
        )
        if test == 0:
            old_beta = beta_smoothing[idx_columns_restricted]
        else:
            old_beta = beta_smoothing

        test += 1
        (
            _,
            idx_columns_restricted,
            _,
            beta_smoothing,
        ) = smoothing_hinge_loss(
            loss_type,
            penalisation_type,
            X_tilde_reduced,
            pairs_reduced,
            alpha,
            old_beta,
            highest_eig,
            tau,
            n_iter,
        )

        if test == 0:
            # ---Dont change samples -> just restrict columns
            X_tilde_reduced = X_tilde_reduced[:, idx_columns_restricted]
            P_reduced = X_tilde_reduced.shape[1]

            highest_eig = power_method(X_tilde_reduced)
            idx_columns = idx_columns_restricted

        # ---Update parameters
        tau = 0.7 * tau
        n_iter = 50

    # ---Results
    beta_smoothing_sample = np.zeros(P)
    for i in range(len(idx_columns)):
        beta_smoothing_sample[idx_columns[i]] = beta_smoothing[i]

    time_smoothing_tot = time.time() - start_time
    LOGGER.info(f"Number of iterations: {test}")
    LOGGER.info("Total time smoothing: " + str(round(time_smoothing_tot, 3)))

    return beta_smoothing_sample


def power_method(X: ndarray, is_sparse=False) -> float:
    """Compute the highest eigenvalue of X^T X.

    Args:
        X (ndarray): The matrix to compute the highest eigenvalue of.
        is_sparse (bool, optional): Whether X is sparse. Defaults to False.

    Returns:
        float: The highest eigenvalue of X^T X.
    """
    P = X.shape[1]

    highest_eigvctr = np.random.rand(P)
    old_highest_eigvctr = -1

    while np.linalg.norm(highest_eigvctr - old_highest_eigvctr) > 1e-2:
        old_highest_eigvctr = highest_eigvctr
        highest_eigvctr = (
            np.dot(X.T, np.dot(X, highest_eigvctr))
            if not is_sparse
            else X.T.dot(X.dot(highest_eigvctr))
        )
        highest_eigvctr /= np.linalg.norm(highest_eigvctr)

    X_highest_eig = (
        np.dot(X, highest_eigvctr) if not is_sparse else X.dot(highest_eigvctr)
    )

    highest_eig = np.dot(X_highest_eig.T, X_highest_eig) / np.linalg.norm(
        highest_eigvctr
    )
    return highest_eig


def soft_thresholding_l1(c: float, alpha: float) -> float:
    """The thresholding operator for L1 penalisation. See Section 4.2 of Dedieu
    et al. (2022).
    """
    if alpha >= abs(c):
        return 0
    else:
        if c >= 0:
            return c - alpha
        else:
            return c + alpha


def soft_thresholding_l2(c: float, alpha: float) -> float:
    """The thresholding operator for L2 penalisation. See Section 4.2 of Dedieu
    et al. (2022).
    """
    return c / float(1 + 2 * alpha)



from rank2plan import Model, Pair, TuningMetric
from rank2plan.lp_models import PrimalLpModel
from rank2plan.lp_models.constraint_column_generation import (
    ConstraintColumnModel,
)
from rank2plan.lp_models.objective_values import compute_main_objective
from rank2plan.metrics import kendall_tau
from rank2plan.lp_models.constraint_column_generation.utils import compute_X_tilde
from pulp import LpSolver
from typing import List
from numpy import ndarray
import numpy as np
import logging
from bayes_opt import BayesianOptimization

LOGGER = logging.getLogger(__name__)


class LpModel(Model):
    def __init__(
        self,
        solver: LpSolver,
        use_column_generation=False,
        use_constraint_generation=False,
        C=1.0,
        tol=1e-4,
    ) -> None:
        """Create a new LP model.

        Args:
            solver (LpSolver): The solver to use

            use_column_generation (bool, optional): Whether to use column
            generation. Defaults to False.

            use_constraint_generation (bool, optional): Whether to use
            constraint generation. Defaults to False.

            C (float, optional): The SVM regularisation parameter. Defaults to
            1.0.

            tol (float, optional): Tolerence value used in constraint or column
            generation for adding constraints and columns. Defaults to 1e-4.

        Raises:
            ValueError: If C is not positive NotImplementedError: For now, using
            only column generation without constraint generation is not
            implemented
        """
        if C <= 0:
            raise ValueError(f"C ({C}) must be positive")
        if not use_column_generation and not use_constraint_generation:
            self._underlying = PrimalLpModel(solver, C=C)
        elif use_constraint_generation:
            self._underlying = ConstraintColumnModel(
                solver, C, tol, no_feature_sampling=(not use_column_generation)
            )
        else:
            raise NotImplementedError("Column generation not implemented yet")

    def tune(
        self,
        X_train: ndarray,
        pairs_train: List[Pair],
        X_val: ndarray,
        pairs_val: List[Pair],
        metric: TuningMetric = TuningMetric.LpMainObjective,
        C_range=(0.001, 100),
        tuning_rounds=25,
    ) -> float:
        if tuning_rounds < 5:
            raise ValueError("tuning_rounds should be at least 5")

        pairs_train = _filter_pairs(X_train, pairs_train)
        pairs_val = _filter_pairs(X_val, pairs_val)
        X_tilde_val = compute_X_tilde(X_val, pairs_val)

        def log_scale(x):
            return np.log10(x)

        def exp_scale(x):
            return 10**x

        def validation_score(weights: ndarray) -> float:
            if metric == TuningMetric.LpMainObjective:
                # negative because we are maximising
                return -compute_main_objective(X_tilde_val, pairs_val, weights)
            elif metric == TuningMetric.KendallTau:
                scores = self._underlying.predict(X_val)
                return kendall_tau(pairs_val, scores)
            else:
                raise ValueError(f"Unknown tuning metric {metric}")

        def f(log_C: float):
            C = exp_scale(log_C)
            self._underlying.refit_with_C_value(C, save_state=True)
            val_score = validation_score(self.weights())  # type: ignore
            LOGGER.info(f"Validation score for C={C}: {val_score}")
            return val_score

        optimiser = BayesianOptimization(
            f=f,
            pbounds={"log_C": (log_scale(C_range[0]), log_scale(C_range[1]))},
            verbose=0,
            random_state=1,
        )

        self._underlying.fit(X_train, pairs_train, save_state=True)
        initial_score = validation_score(self.weights())  # type: ignore
        optimiser.register(
            params={"log_C": log_scale(self._underlying.C)}, target=initial_score
        )
        optimiser.maximize(init_points=5, n_iter=tuning_rounds - 5)

        best_C = exp_scale(optimiser.max["params"]["log_C"])  # type: ignore
        best_score = optimiser.max["target"]  # type: ignore
        LOGGER.info(
            f"Tuning found best C={best_C} with validation score {best_score}",
        )
        self._underlying.state = None
        self._underlying.C = best_C
        return best_C

    def fit(self, X: ndarray, pairs: List[Pair]) -> None:
        """Fit the model.

        Args:
            X (ndarray): The feature matrix

            pairs (List[Pair]): The pairs
        """
        pairs = _filter_pairs(X, pairs)
        self._underlying.fit(X, pairs)

    def predict(self, X):
        return self._underlying.predict(X)

    def weights(self):
        return self._underlying.weights()


def _filter_pairs(X: ndarray, pairs: List[Pair]) -> List[Pair]:
    """Remove pairs where the feature vectors are identical, as they do not
    contribute to the model.

    Args:
        X (ndarray): The feature matrix
        pairs (List[Pair]): The pairs

    Returns:
        List[Pair]: pairs with distinct feature vectors
    """
    return [pair for pair in pairs if not np.array_equal(X[pair.i], X[pair.j])]



from rank2plan import Pair
from typing import List
from numpy import ndarray


def kendall_tau(pairs: List[Pair], scores: ndarray) -> float:
    """Compute the Kendall's tau correlation coefficient between the predicted
    scores and the ranking pairs.

    Args:
        pairs (List[Pair]): The list of ranking pairs, shape (n_pairs,).
        scores (ndarray): The predicted scores, shape (n_samples,).

    Returns:
        float: The Kendall's tau correlation coefficient.
    """
    total = 0
    concordant = 0
    discordant = 0
    for pair in pairs:
        i, j = pair.i, pair.j
        # ignore gaps in kendall tau
        if scores[i] < scores[j]:
            concordant += pair.sample_weight
        elif scores[i] > scores[j]:
            discordant += pair.sample_weight
        total += pair.sample_weight

    return (concordant - discordant) / total



from numpy import ndarray
import numpy as np
from typing import List
from rank2plan import Pair


def compute_main_objective(X_tilde: ndarray, pairs: List[Pair], beta: ndarray) -> float:
    """Compute the main objective of the model, not considering C value.

    Args:
        X_tilde (ndarray): The X_tilde feature matrix
        pairs (List[Pair]): The pairs
        beta (ndarray): The weights

    Returns:
        float: The main objective
    """
    res = 0
    for i, pair in enumerate(pairs):
        res += max(0, pair.sample_weight * pair.gap - X_tilde[i].T.dot(beta))
    return res


def compute_regularisation_objective(beta: ndarray) -> float:
    return np.sum(np.abs(beta))


def compute_overall_objective(
    X_tilde: ndarray, pairs: List[Pair], beta: ndarray, C: float
) -> float:
    return compute_main_objective(
        X_tilde, pairs, beta
    ) * C + compute_regularisation_objective(beta)
